import torch
from torch import nn, Tensor
from torch.nn import functional as F


class ConvModule(nn.Module):
    def __init__(self, c1, c2, k=3, s=1):
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, k, s, k//2, bias=False)
        self.norm = nn.BatchNorm2d(c2)
        self.relu = nn.ReLU6(True)

    def forward(self, x: Tensor) -> Tensor:
        return self.relu(self.norm(self.conv(x)))


def get_link(layer, base_ch, growth_rate):
    if layer == 0:
        return base_ch, 0, []

    link = []
    out_channels = growth_rate

    for i in range(10):
        dv = 2 ** i
        if layer % dv == 0:
            link.append(layer - dv)

            if i > 0: out_channels *= 1.7

    out_channels = int((out_channels + 1) / 2) * 2
    in_channels = 0

    for i in link:
        ch, _, _ = get_link(i, base_ch, growth_rate)
        in_channels += ch

    return out_channels, in_channels, link


class HarDBlock(nn.Module):
    def __init__(self, c1, growth_rate, n_layers):
        super().__init__()
        self.links = []
        layers = []
        self.out_channels = 0

        for i in range(n_layers):
            out_ch, in_ch, link = get_link(i+1, c1, growth_rate)
            self.links.append(link)

            layers.append(ConvModule(in_ch, out_ch))

            if (i % 2 == 0) or (i == n_layers - 1):
                self.out_channels += out_ch

        self.layers = nn.ModuleList(layers)

    def forward(self, x: Tensor) -> Tensor:
        layers = [x]

        for layer in range(len(self.layers)):
            link = self.links[layer]
            tin = []

            for i in link:
                tin.append(layers[i])

            if len(tin) > 1:
                x = torch.cat(tin, dim=1)
            else:
                x = tin[0]

            out = self.layers[layer](x)
            layers.append(out)

        t = len(layers)
        outs = []
        for i in range(t):
            if (i == t - 1) or (i % 2 == 1):
                outs.append(layers[i])

        out = torch.cat(outs, dim=1)
        return out


class FCHarDNet(nn.Module):
    def __init__(self, backbone: str = None, num_classes: int = 19) -> None:
        super().__init__()
        first_ch, ch_list, gr, n_layers = [16, 24, 32, 48], [64, 96, 160, 224, 320], [10, 16, 18, 24, 32], [4, 4, 8, 8, 8]

        self.base = nn.ModuleList([])

        # stem
        self.base.append(ConvModule(3, first_ch[0], 3, 2))
        self.base.append(ConvModule(first_ch[0], first_ch[1], 3))
        self.base.append(ConvModule(first_ch[1], first_ch[2], 3, 2))
        self.base.append(ConvModule(first_ch[2], first_ch[3], 3))

        self.shortcut_layers = []
        skip_connection_channel_counts = []
        ch = first_ch[-1]

        for i in range(len(n_layers)):
            blk = HarDBlock(ch, gr[i], n_layers[i])
            ch = blk.out_channels

            skip_connection_channel_counts.append(ch)
            self.base.append(blk)

            if i < len(n_layers) - 1:
                self.shortcut_layers.append(len(self.base) - 1)

            self.base.append(ConvModule(ch, ch_list[i], k=1))
            ch = ch_list[i]
            
            if i < len(n_layers) - 1:
                self.base.append(nn.AvgPool2d(2, 2))

        prev_block_channels = ch
        self.n_blocks = len(n_layers) - 1

        self.denseBlocksUp = nn.ModuleList([])
        self.conv1x1_up = nn.ModuleList([])

        for i in range(self.n_blocks-1, -1, -1):
            cur_channels_count = prev_block_channels + skip_connection_channel_counts[i]
            blk = HarDBlock(cur_channels_count // 2, gr[i], n_layers[i])
            prev_block_channels = blk.out_channels
            
            self.conv1x1_up.append(ConvModule(cur_channels_count, cur_channels_count//2, 1))
            self.denseBlocksUp.append(blk)

        self.finalConv = nn.Conv2d(prev_block_channels, num_classes, 1, 1, 0)

        self.apply(self._init_weights)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
        elif isinstance(m, nn.BatchNorm2d):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)

    def init_pretrained(self, pretrained: str = None) -> None:
        if pretrained:
            self.load_state_dict(torch.load(pretrained, map_location='cpu'), strict=False)

    def forward(self, x: Tensor) -> Tensor:
        H, W = x.shape[-2:]
        skip_connections = []
        for i, layer in enumerate(self.base):
            x = layer(x)
            if i in self.shortcut_layers:
                skip_connections.append(x)

        out = x

        for i in range(self.n_blocks):
            skip = skip_connections.pop()
            out = F.interpolate(out, size=skip.shape[-2:], mode='bilinear', align_corners=True)
            out = torch.cat([out, skip], dim=1)
            out = self.conv1x1_up[i](out)
            out = self.denseBlocksUp[i](out)

        out = self.finalConv(out)
        out = F.interpolate(out, size=(H, W), mode='bilinear', align_corners=True)
        return out


if __name__ == '__main__':
    model = FCHarDNet()
    # model.init_pretrained('checkpoints/backbones/hardnet/hardnet_70.pth')
    # model.load_state_dict(torch.load('checkpoints/pretrained/hardnet/hardnet70_cityscapes.pth', map_location='cpu'))
    x = torch.zeros(1, 3, 224, 224)
    outs = model(x)
    print(outs.shape)



import pandas as pd
import wandb

from ezdl.utils.grid import linearized_to_string
from ezdl.experiment.experiment import Experimenter


def format_bool(value: bool):
    if value:
        return '<span style="color:green">True</span>'
    else:
        return '<span style="color:red">False</span>'


def exp_summary_builder(exp: Experimenter):
    txt = ""
    if exp.gs is None:
        return txt
    txt += "### Experiment summary \n"
    txt += "|Property | Value |\n"
    txt += "|---------|-------|\n"
    txt += f"|Resume experiment            | {format_bool(exp.exp_settings.resume)} |\n"
    txt += f"|Resume interrupted run       | {format_bool(exp.exp_settings.resume_last)} |\n"
    txt += f"|Continue with errors         | {format_bool(exp.exp_settings.continue_with_errors)} |\n"
    txt += f"|Excluded files               | {exp.exp_settings.excluded_files} |\n"
    txt += f"|Tracking directory           | {exp.exp_settings.tracking_dir} |\n"
    txt += f"|Starting grid                | {exp.exp_settings.start_from_grid} |\n"
    txt += f"|Starting run                 | {exp.exp_settings.start_from_run} |\n"
    txt += f"|Total runs                   | {exp.gs.total_runs} |\n"
    txt += f"|Total runs to run            | {exp.gs.total_runs_to_run} |\n"
    txt += f"|Total runs excluded by grids | {exp.gs.total_runs_excl_grid} |\n"
    txt += f"|Total runs excluded          | {exp.gs.total_runs_excl} |\n"

    return txt


def grid_summary_builder(grids, dot_elements):
    dfs = [pd.DataFrame(linearized_to_string(dot_element), columns=[f"Grid {i}", f"N. runs: {len(grid)}"])
           for i, (dot_element, grid) in enumerate(zip(dot_elements, grids))]
    mark_grids = "\n\n".join(df.to_markdown(index=False) for df in dfs)
    mark_grids = "### Most important parameters for each grid \n" + mark_grids
    return mark_grids


def format_to_delete(keys, files, artifacts):
    pass


def title_builder(title):
    return f"## {title}"


def wandb_run_link(wandb_run):
    return f"View run on wandb: [{wandb_run.name}]({wandb_run.url})"


class MkFailures:
    def __init__(self):
        self.value = "## Failures \n\n"

    def get_text(self):
        return self.value

    def update(self, cur_grid, cur_run, exception):
        self.value += f"Grid {cur_grid} Run {cur_run} failed with an exception: \n\n"
        self.value += f"{exception} \n\n --- \n\n"



from .resnet import ResNet, resnet_settings
from .resnetd import ResNetD, resnetd_settings
from .micronet import MicroNet, micronet_settings
from .mobilenetv2 import MobileNetV2, mobilenetv2_settings
from .mobilenetv3 import MobileNetV3, mobilenetv3_settings

from .mit import MiT, mit_settings
from .pvt import PVTv2, pvtv2_settings
from .rest import ResT, rest_settings
from .poolformer import PoolFormer, poolformer_settings
from .convnext import ConvNeXt, convnext_settings


__all__ = [
    'ResNet', 
    'ResNetD', 
    'MicroNet',
    'MobileNetV2',
    'MobileNetV3',
    
    'MiT', 
    'PVTv2', 
    'ResT',
    'PoolFormer',
    'ConvNeXt',
]


from torchvision.models.segmentation.deeplabv3 import deeplabv3_resnet50 as dls50


def deeplabv3_resnet50(arch_params):
    args = dls50.__code__.co_varnames
    reduced_keys = set(arch_params.keys()).intersection(args)
    reduced_args = {k: arch_params[k] for k in reduced_keys if k in arch_params}
    return dls50(**reduced_args)



import numpy as np
from matplotlib import colormaps
import plotly.express as px


class ColorMap:
    def __init__(self):
        self.cmap = ["#000000", "#00ff00", "#ff0000", "#0000ff"] \
                    + px.colors.qualitative.Alphabet
        self.cmap = [tuple(int(h.lstrip('#')[i:i+2], 16) for i in (0, 2, 4)) for h in self.cmap]

    def __getitem__(self, item):
        return self.cmap[item]


def tensor_to_segmentation_image(prediction, cmap: list = None, labels=None, return_cmap=False) -> np.array:
    if cmap is None:
        cmap = ColorMap()
    if labels is None:
        labels = np.unique(prediction)
    segmented_image = np.ones((*prediction.shape, 3), dtype="uint8")
    for i in range(len(labels)):
        segmented_image[prediction == i] = cmap[i]
    if return_cmap:
        cmap = {labels[i]: cmap[i] for i in range(len(labels))}
        return segmented_image, cmap
    return segmented_image



import gc
import sys
import os

from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.training.utils.callbacks import Phase
from copy import deepcopy

from ezdl.callbacks import MetricsLogCallback, callback_factory
from ezdl.experiment.parameters import parse_params
from ezdl.experiment.seg_trainer import SegmentationTrainer
from ezdl.utils.utilities import dict_to_yaml_string, values_to_number, nested_dict_update

logger = get_logger(__name__)


class Run:
    def __init__(self):
        self.params = None
        self.dataset = None
        self.train_callbacks = None
        self.val_callbacks = None
        self.test_callbacks = None
        self.dataset_params = None
        self.seg_trainer = None
        self.train_params = None
        self.test_params = None
        self.run_params = None
        self.phases = None
        if '.' not in sys.path:
            sys.path.extend('.')

    def parse_params(self, params):
        self.params = deepcopy(params)
        self.phases = params['phases']

        self.train_params, self.test_params, self.dataset_params, callbacks = parse_params(self.params)
        self.train_callbacks, self.val_callbacks, self.test_callbacks = callbacks
        self.run_params = params.get('run_params') or {}

    def init(self, params: dict):
        self.seg_trainer = None
        try:
            self.parse_params(params)
            self.seg_trainer = SegmentationTrainer(
                experiment_name=self.params['experiment']['group'],
                ckpt_root_dir=self.params['experiment']['tracking_dir'] or 'wandb',
            )
            self.dataset = self.seg_trainer.init_dataset \
                    (params['dataset_interface'], dataset_params=deepcopy(self.dataset_params))
            self.seg_trainer.init_model(params, False, None)
            self.seg_trainer.init_loggers({"in_params": params}, deepcopy(self.train_params))
            logger.info(f"Input params: \n\n {dict_to_yaml_string(params)}")
        except Exception as e:
            if (
                self.seg_trainer is not None
                and self.seg_trainer.sg_logger is not None
            ):
                self.seg_trainer.sg_logger.close(True)
            raise e

    def resume(self, wandb_run, updated_config, phases):
        try:
            try:
                self.params = values_to_number(wandb_run.config['in_params'])
            except KeyError as e:
                raise RuntimeError("No params recorded for run, just delete it!") from e
            self.params = nested_dict_update(self.params, updated_config)
            self.phases = phases
            wandb_run.config['in_params'] = self.params
            wandb_run.update()
            self.train_params, self.test_params, self.dataset_params, callbacks = parse_params(self.params)
            self.train_callbacks, self.val_callbacks, self.test_callbacks = callbacks

            self.seg_trainer = SegmentationTrainer(
                experiment_name=self.params['experiment']['group'],
                ckpt_root_dir=self.params['experiment']['tracking_dir'] or 'wandb',
            )
            self.dataset = self.seg_trainer.init_dataset \
                (wandb_run.config['in_params']['dataset_interface'], dataset_params=deepcopy(self.dataset_params))
            track_dir = wandb_run.config.get('in_params').get('experiment').get('tracking_dir') or 'wandb'
            checkpoint_path_group = os.path.join(track_dir, wandb_run.group, 'wandb')
            run_folder = list(filter(lambda x: str(wandb_run.id) in x, os.listdir(checkpoint_path_group)))
            checkpoint_path = None
            if 'epoch' in wandb_run.summary:
                ckpt = 'ckpt_latest.pth' if 'train' in phases else 'ckpt_best.pth'
                try:
                    checkpoint_path = os.path.join(checkpoint_path_group, run_folder[0], 'files', ckpt)
                except IndexError as exc:
                    logger.error(f"{wandb_run.id} not found in {checkpoint_path_group}")
                    raise ValueError(
                        f"{wandb_run.id} not found in {checkpoint_path_group}"
                    ) from exc
            self.seg_trainer.init_model(self.params, True, checkpoint_path)
            self.seg_trainer.init_loggers({"in_params": self.params}, self.train_params, run_id=wandb_run.id)
        except Exception as e:
            if self.seg_trainer is not None:
                self.seg_trainer.sg_logger.close(really=True)
            raise e

    def launch(self):
        try:
            if 'train' in self.phases:
                train(self.seg_trainer, self.train_params, self.dataset, self.train_callbacks, self.val_callbacks)

            if 'test' in self.phases:
                test_metrics = self.seg_trainer.test(**self.test_params, test_phase_callbacks=self.test_callbacks)

            if 'inference' in self.phases:
                inference(self.seg_trainer, self.run_params, self.dataset)
        finally:
            if self.seg_trainer is not None:
                self.seg_trainer.sg_logger.close(True)


def train(seg_trainer, train_params, dataset, train_callbacks, val_callbacks):
    # Callbacks

    cbcks = [callback_factory(name, params, seg_trainer=seg_trainer, dataset=dataset, loader=dataset.train_loader)
             for name, params in {**train_callbacks, **val_callbacks}.items()
             ]
    cbcks = [
        MetricsLogCallback(Phase.TRAIN_EPOCH_END, freq=1),
        MetricsLogCallback(Phase.VALIDATION_EPOCH_END, freq=1),
        *cbcks
    ]
    train_params["phase_callbacks"] = cbcks

    seg_trainer.train(train_params)
    gc.collect()


def inference(seg_trainer, run_params, dataset):
    run_loader = dataset.get_run_loader(folders=run_params['run_folders'], batch_size=run_params['batch_size'])
    cbcks = [
        # SaveSegmentationPredictionsCallback(phase=Phase.POST_TRAINING,
        #                                     path=
        #                                     run_params['prediction_folder']
        #                                     if run_params['prediction_folder'] != 'mlflow'
        #                                     else mlclient.run.info.artifact_uri + '/predictions',
        #                                     num_classes=len(seg_trainer.test_loader.dataset.classes),
        #                                     )
    ]
    run_loader.dataset.return_name = True
    seg_trainer.run(run_loader, callbacks=cbcks)
    # seg_trainer.valid_loader.dataset.return_name = True
    # seg_trainer.run(seg_trainer.valid_loader, callbacks=cbcks)



import numpy as np
import torch
from super_gradients.training.utils.callbacks import Phase
from super_gradients.training.utils.early_stopping import EarlyStop

from ezdl.loss import LOSSES as LOSSES_DICT
from ezdl.metrics import metrics_factory
from ezdl.utils.utilities import recursive_get


def parse_params(params: dict) -> (dict, dict, dict, list):
    # Set Random seeds
    torch.manual_seed(params['train_params']['seed'])
    np.random.seed(params['train_params']['seed'])

    # Instantiate loss
    input_train_params = params['train_params']
    loss_params = params['train_params']['loss']
    loss = LOSSES_DICT[loss_params['name']](**loss_params['params'])

    # metrics
    train_metrics = metrics_factory(params['train_metrics'])
    test_metrics = metrics_factory(params['test_metrics'])

    # dataset params
    dataset_params = params['dataset']

    if input_train_params.get('metric_to_watch') == 'loss':
        input_train_params['metric_to_watch'] = loss.__class__.__name__

    train_params = {
        **input_train_params,
        "train_metrics_list": list(train_metrics.values()),
        "valid_metrics_list": list(test_metrics.values()),
        "loss": loss,
        "loss_logging_items_names": ["loss"],
        "sg_logger": params['experiment']['logger'],
        'sg_logger_params': {
            'entity': params['experiment']['entity'],
            'tags': params['tags'],
            'project_name': params['experiment']['name'],
        }
    }

    test_params = {
        "test_metrics": test_metrics,
    }

    # callbacks
    train_callbacks = add_phase_in_callbacks(params.get('train_callbacks') or {}, "train")
    test_callbacks = add_phase_in_callbacks(params.get('test_callbacks') or {}, "test")
    val_callbacks = add_phase_in_callbacks(params.get('val_callbacks') or {}, "validation")

    # early stopping
    if params.get('early_stopping'):
        val_callbacks['early_stopping'] = params.get('early_stopping')

    if recursive_get(val_callbacks, 'early_stopping', 'monitor') == 'loss':
        val_callbacks['early_stopping']['monitor'] = loss.__class__.__name__

    return train_params, test_params, dataset_params, (train_callbacks, val_callbacks, test_callbacks)


def add_phase_in_callbacks(callbacks, phase):
    """
    Add default phase to callbacks
    :param callbacks: dict of callbacks
    :param phase: "train", "validation" or "test"
    :return: dict of callbacks with phase
    """
    for callback in callbacks.values():
        if callback.get('phase') is None:
            callback['phase'] = phase
    return callbacks



import torch
from torch import Tensor
from torch.nn import functional as F
from ezdl.models.base import BaseModel
from ezdl.models.heads import SegFormerHead


class SegFormer(BaseModel):
    def __init__(self, backbone: str = 'MiT-B0', num_classes: int = 19) -> None:
        super().__init__(backbone, num_classes)
        self.decode_head = SegFormerHead(self.backbone.channels, 256 if 'B0' in backbone or 'B1' in backbone else 768, num_classes)
        self.apply(self._init_weights)

    def forward(self, x: Tensor) -> Tensor:
        y = self.backbone(x)
        y = self.decode_head(y)   # 4x reduction in image size
        y = F.interpolate(y, size=x.shape[2:], mode='bilinear', align_corners=False)    # to original image shape
        return y


if __name__ == '__main__':
    model = SegFormer('MiT-B0')
    # model.load_state_dict(torch.load('checkpoints/pretrained/segformer/segformer.b0.ade.pth', map_location='cpu'))
    x = torch.zeros(1, 3, 512, 512)
    y = model(x)
    print(y.shape)


import streamlit as st
import wandb
import os

from ruamel.yaml.scanner import ScannerError
from streamlit_ace import st_ace
from streamlit.web import bootstrap
from streamlit import config as _config

from ezdl.app.utils import ManipulationInputs
from ezdl.app.markdown import exp_summary_builder, grid_summary_builder, MkFailures, wandb_run_link, format_to_delete
from ezdl.experiment.experiment import ExpSettings, Experimenter, Status

from ezdl.utils.utilities import load_yaml, dict_to_yaml, yaml_string_to_dict
from ezdl.wandb_manip import update_metadata, fix_string_param, remove_key, delete_files, delete_artifacts, update_config

STREAMLIT_AGGRID_URL = "https://github.com/PablocFonseca/streamlit-aggrid"


def DEFAULTS(key=None):
    if key is None:
        return "", "", "", "", "", 0, 0, []
    if isinstance(key, dict):
        return {v: DEFAULTS(k) for k, v in key.items()}
    if isinstance(key, list):
        return {k: DEFAULTS(k) for k in key}
    if key in ['start_from_grid', 'start_from_run']:
        return 0
    elif key == "flags":
        return []
    else:
        return ""


def update_from_gui(group, track_dir, start_grid, start_run, resume, continue_errors):
    if st.session_state.experimenter is None:
        st.session_state.experimenter = Experimenter()
    st.session_state.experimenter.update_settings({
        "resume": resume,
        "continue_with_errors": continue_errors,
        "start_from_grid": start_grid,
        "start_from_run": start_run,
        "tracking_dir": track_dir,
        "group": group,
    })
    st.session_state.mk_summary = exp_summary_builder(st.session_state.experimenter)


def set_file():
    st.session_state.experimenter = Experimenter()
    st.session_state.experimenter.exp_settings = ExpSettings()
    st.session_state.mk_summary = ""
    st.session_state.mk_params = ""
    file = st.session_state.parameter_file
    if file is None:
        return
    settings, st_string = load_yaml(file, return_string=True)
    st.session_state.settings_string = st_string
    set_settings(settings)


def edit_file():
    st.session_state.experimenter = Experimenter()
    st.session_state.experimenter.exp_settings = ExpSettings()
    st.session_state.mk_summary = ""
    st.session_state.mk_params = ""
    settings = yaml_string_to_dict(st.session_state.settings_string)
    set_settings(settings)


def set_settings(settings):
    sm, grids, dots = st.session_state.experimenter.calculate_runs(settings)
    st.session_state.mk_summary = exp_summary_builder(st.session_state.experimenter)
    st.session_state.mk_params = grid_summary_builder(grids, dots)


class Interface:
    def __init__(self, parameter_file, exp_settings: ExpSettings = None, share: bool = True):
        self.mk_wandb_link = None
        self.mk_cur_params = None
        self.failures = None
        self.mk_failures = None
        self.current_params = None
        if "experimenter" not in st.session_state:
            self.experimenter = Experimenter()
            st.session_state.experimenter = self.experimenter
            self.experimenter.exp_settings.update(exp_settings)
            st.session_state.mk_summary = exp_summary_builder(self.experimenter)
        else:
            self.experimenter = st.session_state.experimenter

        st.set_page_config(
            layout="wide", page_icon="🖱️", page_title="EzDL"
        )
        st.title("EzDL")
        tab1, tab2, = st.tabs(["Training", "Manipulation"])
        with tab1:
            self.experiment_interface()
        with tab2:
            self.manipulation_interface()

    def experiment_interface(self):
        st.write(
            """Train your model!"""
        )
        es = self.experimenter.exp_settings
        with st.sidebar:
            with st.form("exp_form"):
                self.form_path = st.text_input("Experiment name (path)", value=es.name)
                self.form_group = st.text_input("Experiment group", value=es.group)
                self.form_track_dir = st.text_input("Tracking directory", value=es.tracking_dir)
                self.form_grid = st.number_input("Start from grid", min_value=0, value=es.start_from_grid)
                self.form_run = st.number_input("Start from run", min_value=0, value=es.start_from_run)
                self.form_resume = st.checkbox("Resume", value=es.resume)
                self.form_continue = st.checkbox("Continue with errors", value=es.continue_with_errors)
                submitted = st.form_submit_button("Submit")
                if submitted:
                    update_from_gui(group=self.form_group,
                                    track_dir=self.form_track_dir,
                                    start_grid=self.form_grid,
                                    start_run=self.form_run,
                                    resume=self.form_resume,
                                    continue_errors=self.form_continue)

            self.parameter_file = st.file_uploader("Parameter file", on_change=set_file, key="parameter_file")
        if "parameter_file" in st.session_state and st.session_state.parameter_file:
            st.write("## ", f"{es.name} - {es.group}")
            self.yaml_error = st.empty()
            with st.expander("Grid file"):
                st.session_state.edit_mode = st.button("Edit") ^ \
                                             ("edit_mode" in st.session_state and st.session_state.edit_mode)
                if st.session_state.edit_mode:
                    st.session_state.settings_string = st_ace(
                        value=st.session_state.settings_string, language="yaml", theme="twilight")
                    try:
                        edit_file()
                    except ScannerError as e:
                        self.yaml_error.exception(e)
                else:
                    st.code(st.session_state.settings_string, language="yaml")
            col1, col2 = st.columns(2)
            with col1:
                st.markdown(st.session_state.mk_summary, unsafe_allow_html=True)
            with col2:
                st.write(st.session_state.mk_params)
            launch = st.button("Launch!")
            if launch:
                self.experiment()

    def manipulation_interface(self):
        path = st.text_input("wandb experiment path")
        col1, col2 = st.columns(2)
        with col1:
            st.text("Filters")
            filters = st_ace(language="yaml", theme="twilight", key="filters")
        with col2:
            st.text("Updated config")
            updated_config = st_ace(language="yaml", theme="twilight", key="updated_config")
        col1, col2 = st.columns(2)
        with col1:
            st.text("Updated meta config")
            updated_meta = st_ace(language="yaml", theme="twilight", key="updated_meta")
        with col2:
            st.text("What to delete")
            to_delete_place = "keys:\n" \
                              "files:\n" \
                              "artifacts:\n"
            to_delete = st_ace(value=to_delete_place, language="yaml", theme="twilight", key="to_delete")

        fix_string_params = st.checkbox("Fix string parameters turning them into numbers")

        col1, col2 = st.columns(2)
        with col1:
            launch = st.button("Go")
        with col2:
            preview = st.button("Retrieve runs and preview")

        manip = ManipulationInputs()
        manip.update(
            path, filters, updated_config, updated_meta, to_delete, fix_string_params
        )

        if "manip" not in st.session_state:
            st.session_state.manip = manip

        if launch:
            self.manipulate()

        if preview:
            self.preview_manipulation()

    def manipulate(self):
        mp = st.session_state.manip
        api = wandb.Api()
        runs = api.runs(path=mp.path, filters=mp.filters)
        if len(runs) != 0:
            for run in runs:
                print(f"Name: {run.name} Id: {run.id}")
                if mp.fix_string_params:
                    fix_string_param(run)
                if mp.updated_config is None:
                    print('No config to update')
                else:
                    update_config(run, mp.updated_config)
                if mp.updated_metadata is None:
                    print('No metadata to update')
                else:
                    update_metadata(run, mp.updated_metadata)
                if mp.keys_to_delete is None:
                    print('No keys to delete')
                else:
                    remove_key(run, mp.keys_to_delete)
                if mp.files_to_delete is None:
                    print("No files to delete")
                else:
                    delete_files(run, mp.files_to_delete)
                if mp.artifacts_to_delete is None:
                    print("No artifacts to delete")
                else:
                    delete_artifacts(run, mp.artifacts_to_delete)

    def preview_manipulation(self):
        mp = st.session_state.manip
        with open("log.txt") as log:
            log.write(f"Filters: {mp.filters} path: {mp.path}")
        api = wandb.Api()
        runs = api.runs(path=mp.path, filters=mp.filters)
        col1, col2 = st.columns([2, 1])
        with col1:
            st.table(
                {"Name": list(map(lambda x: x.name, runs)), "Group": list(map(lambda x: x.group, runs)), "Id": list(map(lambda x: x.id, runs))}
            )
        with col2:
            st.markdown(format_to_delete(mp.keys_to_delete, mp.files_to_delete, mp.artifacts_to_delete))

    def update_bars(self, bars, status: Status):
        for i in range(status.n_grids):
            if i < status.grid:
                bars[i].progress(1.0)
            elif i == status.grid:
                if status in [Status.FINISHED, Status.CRASHED]:
                    status.run += 1
                bars[i].progress(status.run / status.grid_len)
            else:
                bars[i].progress(0)
        self.current_params.code(dict_to_yaml(status.params), language="yaml")
        self.mk_progress.text(f"Grid {status.grid} / {status.n_grids - 1} \n"
                              f"Run  {status.run} / {status.grid_len - 1}")
        if status == "crashed":
            self.failures.update(status.grid, status.run, status.exception)
            self.mk_failures.markdown(self.failures.get_text())
        if status == Status.STARTING:
            print("Just started")
        if status.wandb_run is not None:
            self.mk_wandb_link.markdown(wandb_run_link(status.wandb_run))

    def experiment(self):
        self.mk_progress = st.text("Starting... wait!")
        bars = [st.progress(0) for i in range(len(st.session_state.experimenter.grids))]
        col1, col2 = st.columns(2)
        with col1:
            self.mk_cur_params = st.markdown("### Current run")
            self.mk_wandb_link = st.markdown("Waiting to create run on Wandb")
            self.current_params = st.empty()
        with col2:
            self.failures = MkFailures()
            self.mk_failures = st.markdown("### Failures")
        for status in st.session_state.experimenter.execute_runs_generator():
            self.update_bars(bars, status)
        st.balloons()


def launch_streamlit(args):
    cli_args = ['--grid', str(args.grid),
                '--run', str(args.run),
                ]
    if args.dir:
        cli_args += ['--dir', args.dir]
    if args.file:
        cli_args += ['--file', args.file]
    if args.resume:
        cli_args += ['--resume']

    dirname = os.path.dirname(__file__)
    filename = os.path.join(dirname, 'guicli.py')

    _config.set_option("server.headless", True)
    print(cli_args)
    bootstrap.run(filename, '', cli_args, flag_options={})


def streamlit_entry(parameter_file, args, share):
    settings = ExpSettings(**args)
    Interface(parameter_file, settings, share)


def frontend(args):
    launch_streamlit(args)



import torch
import math
from torch import nn, Tensor
from torch.nn import functional as F
from ezdl.models.layers import ConvModule


class DetailBranch(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.S1 = nn.Sequential(
            ConvModule(3, 64, 3, 2, 1),
            ConvModule(64, 64, 3, 1, 1)
        )
        self.S2 = nn.Sequential(
            ConvModule(64, 64, 3, 2, 1),
            ConvModule(64, 64, 3, 1, 1),
            ConvModule(64, 64, 3, 1, 1)
        )
        self.S3 = nn.Sequential(
            ConvModule(64, 128, 3, 2, 1),
            ConvModule(128, 128, 3, 1, 1),
            ConvModule(128, 128, 3, 1, 1)
        )

    def forward(self, x):
        return self.S3(self.S2(self.S1(x)))


class StemBlock(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.conv_3x3 = ConvModule(3, 16, 3, 2, 1)
        self.left = nn.Sequential(
            ConvModule(16, 8, 1, 1, 0),            
            ConvModule(8, 16, 3, 2, 1)
        )
        self.right = nn.MaxPool2d(3, 2, 1, ceil_mode=False)
        self.fuse = ConvModule(32, 16, 3, 1, 1)

    def forward(self, x):
        x = self.conv_3x3(x)
        x_left = self.left(x)
        x_right = self.right(x)
        y = torch.cat([x_left, x_right], dim=1)
        return self.fuse(y)


class ContextEmbeddingBlock(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.inner = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            nn.BatchNorm2d(128),
            ConvModule(128, 128, 1, 1, 0)           
        )
        self.conv = ConvModule(128, 128, 3, 1, 1)   

    def forward(self, x):
        y = self.inner(x)
        out = x + y
        return self.conv(out)


class GatherExpansionLayerv1(nn.Module):
    def __init__(self, in_ch, out_ch, e=6) -> None:
        super().__init__()
        self.inner = nn.Sequential(
            ConvModule(in_ch, in_ch, 3, 1, 1),
            ConvModule(in_ch, in_ch*e, 3, 1, 1, g=in_ch),  
            nn.Conv2d(in_ch*e, out_ch, 1, bias=False),
            nn.BatchNorm2d(out_ch)
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        y = self.inner(x)
        out = x + y
        return self.relu(out)


class GatherExpansionLayerv2(nn.Module):
    def __init__(self, in_ch, out_ch, e=6) -> None:
        super().__init__()
        self.inner = nn.Sequential(
            ConvModule(in_ch, in_ch, 3, 1, 1),
            nn.Conv2d(in_ch, in_ch*e, 3, 2, 1, groups=in_ch, bias=False),
            nn.BatchNorm2d(in_ch*e),
            ConvModule(in_ch*e, in_ch*e, 3, 1, 1, g=in_ch*e),               
            nn.Conv2d(in_ch*e, out_ch, 1, bias=False),
            nn.BatchNorm2d(out_ch)
        )

        self.outer = nn.Sequential(
            nn.Conv2d(in_ch, in_ch, 3, 2, 1, groups=in_ch, bias=False),
            nn.BatchNorm2d(in_ch),
            nn.Conv2d(in_ch, out_ch, 1, bias=False),
            nn.BatchNorm2d(out_ch)
        )

        self.relu = nn.ReLU(True)

    def forward(self, x):
        x1 = self.inner(x)
        x2 = self.outer(x)
        out = x1 + x2
        return self.relu(out)


class SemanticBranch(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.S1S2 = StemBlock()
        self.S3 = nn.Sequential(
            GatherExpansionLayerv2(16, 32),
            GatherExpansionLayerv1(32, 32)
        )
        self.S4 = nn.Sequential(
            GatherExpansionLayerv2(32, 64),
            GatherExpansionLayerv1(64, 64)
        )
        self.S5_1 = nn.Sequential(
            GatherExpansionLayerv2(64, 128),
            GatherExpansionLayerv1(128, 128),
            GatherExpansionLayerv1(128, 128),
            GatherExpansionLayerv1(128, 128)
        )
        self.S5_2 = ContextEmbeddingBlock()

    def forward(self, x):
        x2 = self.S1S2(x)
        x3 = self.S3(x2)
        x4 = self.S4(x3)
        x5_1 = self.S5_1(x4)
        x5_2 = self.S5_2(x5_1)
        return x2, x3, x4, x5_1, x5_2


class AggregationLayer(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.left1 = nn.Sequential(
            nn.Conv2d(128, 128, 3, 1, 1, groups=128, bias=False),
            nn.BatchNorm2d(128),
            nn.Conv2d(128, 128, 1, 1, 0, bias=False)
        )
        self.left2 = nn.Sequential(
            nn.Conv2d(128, 128, 3, 2, 1, bias=False),
            nn.BatchNorm2d(128),
            nn.AvgPool2d(3, 2, 1, ceil_mode=False)
        )

        self.right1 = nn.Sequential(
            nn.Conv2d(128, 128, 3, 1, 1, bias=False),
            nn.BatchNorm2d(128),
            nn.Upsample(scale_factor=4),
            nn.Sigmoid()
        )
        self.right2 = nn.Sequential(
            nn.Conv2d(128, 128, 3, 1, 1, groups=128, bias=False),
            nn.BatchNorm2d(128),
            nn.Conv2d(128, 128, 1, 1, 0, bias=False),
            nn.Sigmoid()
        )

        self.up = nn.Upsample(scale_factor=4)           
        self.conv = ConvModule(128, 128, 3, 1, 1)       


    def forward(self, x_d, x_s):
        x1 = self.left1(x_d)
        x2 = self.left2(x_d)
        x3 = self.right1(x_s)
        x4 = self.right2(x_s)

        left = x1 * x3
        right = x2 * x4
        right = self.up(right)
        out = left + right

        return self.conv(out)


class SegHead(nn.Module):
    def __init__(self, in_ch, mid_ch, num_classes, upscale_factor=8, is_aux=True) -> None:
        super().__init__()
        out_ch = num_classes * upscale_factor * upscale_factor

        self.conv_3x3 = ConvModule(in_ch, mid_ch, 3, 1, 1)
        self.drop = nn.Dropout(0.1)

        if is_aux:
            self.conv_out = nn.Sequential(
                ConvModule(mid_ch, upscale_factor * upscale_factor, 3, 1, 1),
                nn.Conv2d(upscale_factor*upscale_factor, out_ch, 1, 1, 0),
                nn.PixelShuffle(upscale_factor)
            )
        else:
            self.conv_out = nn.Sequential(
                nn.Conv2d(mid_ch, out_ch, 1, 1, 0),
                nn.PixelShuffle(upscale_factor)
            )

    def forward(self, x):
        out = self.conv_3x3(x)
        out = self.drop(out)
        return self.conv_out(out)


class BiSeNetv2(nn.Module):
    def __init__(self, backbone: str = None, num_classes: int = 19) -> None:
        super().__init__()
        self.detail_branch = DetailBranch()
        self.semantic_branch = SemanticBranch()
        self.aggregation_layer = AggregationLayer()

        self.output_head = SegHead(128, 1024, num_classes, upscale_factor=8, is_aux=False)

        self.aux2_head = SegHead(16, 128, num_classes, upscale_factor=4)
        self.aux3_head = SegHead(32, 128, num_classes, upscale_factor=8)
        self.aux4_head = SegHead(64, 128, num_classes, upscale_factor=16)
        self.aux5_head = SegHead(128, 128, num_classes, upscale_factor=32)

        self.apply(self._init_weights)

    def _init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Conv2d):
            fan_out = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            fan_out // m.groups
            m.weight.data.normal_(0, math.sqrt(2.0 / fan_out))
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm2d)):
            nn.init.ones_(m.weight)
            nn.init.zeros_(m.bias)

    def init_pretrained(self, pretrained: str = None) -> None:
        pass

    def forward(self, x):
        x_d = self.detail_branch(x)
        aux2, aux3, aux4, aux5, x_s = self.semantic_branch(x)

        output = self.aggregation_layer(x_d, x_s)
        output = self.output_head(output)

        if self.training:
            aux2 = self.aux2_head(aux2)
            aux3 = self.aux3_head(aux3)
            aux4 = self.aux4_head(aux4)
            aux5 = self.aux5_head(aux5)
            return output, aux2, aux3, aux4, aux5
        return output


if __name__ == '__main__':
    model = BiSeNetv2()
    model.eval()
    image = torch.randn((1, 3, 224, 224))
    output = model(image)
    print(output.shape)


from typing import Mapping
from functools import reduce

from torch import Tensor
from torchmetrics import JaccardIndex, AUROC, F1Score, Precision, Recall, ConfusionMatrix
from torchmetrics.functional.classification.roc import _roc_compute
import torch

from copy import deepcopy


class AUC(AUROC):
    def update(self, preds: Tensor, target: Tensor) -> None:
        AUROC.update(self, preds.cpu(), target.cpu())

    def get_roc(self):
        preds = torch.cat(self.preds, dim=0)
        target = torch.cat(self.target, dim=0)
        if not self.num_classes:
            raise ValueError(f"`num_classes` bas to be positive number, but got {self.num_classes}")
        return _roc_compute(preds, target, self.num_classes, self.pos_label)


def PerClassAUC(name, code):
    def __init__(self, name, code, *args, **kwargs):
        AUC.__init__(self, **kwargs)
        self.code = code

    def update(self, preds: Tensor, target: Tensor) -> None:
        preds = preds[:, self.code, ::].flatten()
        target = (target == code).flatten()
        AUC.update(self, preds, target)

    metric = type(name, (AUC,), {"update": update, "__init__": __init__})
    return metric(name, code)


# class WrapJaccard(JaccardIndex):
#     def update(self, preds: Tensor, target: Tensor) -> None:
#         target = target.argmax(dim=1) if len(target.shape) > 1 else target
#         return super().update(preds.argmax(dim=1), target)
#
#
# class WrapF1(F1Score):
#     def update(self, preds: Tensor, target: Tensor) -> None:
#         return super().update(preds.argmax(dim=1), target.argmax(dim=1))
#
#
# class WrapPrecision(Precision):
#     def update(self, preds: Tensor, target: Tensor) -> None:
#         target = target.argmax(dim=1) if len(target.shape) > 1 else target
#         return super().update(preds.argmax(dim=1), target)
#
#
# class WrapRecall(Recall):
#     def update(self, preds: Tensor, target: Tensor) -> None:
#         target = target.argmax(dim=1) if len(target.shape) > 1 else target
#         return super().update(preds.argmax(dim=1), target)


class WrapCF(ConfusionMatrix):
    # def update(self, preds: Tensor, target: Tensor) -> None:
    #     return super().update(preds.argmax(dim=1), target.argmax(dim=1))

    def compute(self) -> Tensor:
        # PlaceHolder value
        return Tensor([1])

    def get_cf(self):
        return super().compute()


def metric_instance(name: str, params: dict) -> dict:
    if params.get('discriminator') is not None:
        params = deepcopy(params)
        names = params.pop('discriminator')
        return {
            subname: METRICS[name](subname, code, **params)
            for subname, code in names
        }
    return {name: METRICS[name](**params)}


def metrics_factory(metrics_params: Mapping) -> dict:
    return reduce(lambda a, b: {**a, **b},
                  [
                      metric_instance(name, params)
                      for name, params in metrics_params.items()
                  ]
                  )


METRICS = {
    'jaccard': JaccardIndex,
    'auc': AUC,
    'perclassauc': PerClassAUC,
    'f1': F1Score,
    'precision': Precision,
    'recall': Recall,
    'conf_mat': WrapCF
}



import time

import gradio as gr
import numpy as np
import pandas as pd
import torch
import wandb

from ezdl.app.markdown import grid_summary_builder, exp_summary_builder, title_builder, MkFailures
from ezdl.utils.utilities import load_yaml, update_collection
from ezdl.utils.segmentation import tensor_to_segmentation_image
from ezdl.app.inference import WandbInferencer, compose_images

from ezdl.experiment.experiment import Experimenter, ExpSettings

DEFAULT_CMAP = {
    0: (0, 0, 0),
    1: (255, 0, 0),
    2: (0, 255, 0)
}


def DEFAULTS(key=None):
    if key is None:
        return "", "", "", "", "", 0, 0, []
    if isinstance(key, dict):
        return {v: DEFAULTS(k) for k, v in key.items()}
    if isinstance(key, list):
        return {k: DEFAULTS(k) for k in key}
    if key in ['start_from_grid', 'start_from_run']:
        return 0
    elif key == "flags":
        return []
    else:
        return ""


TITLE = "# EzDL"
TEXT = "Load config and model to select image channels"
DIV = "---"
COLS = ['Grid', 'N. runs']


def segment(pred: torch.Tensor):
    pred = pred.squeeze(0).argmax(dim=0).cpu()
    return tensor_to_segmentation_image(pred, DEFAULT_CMAP)


def flag_maker(exp_setings: ExpSettings):
    flags = []
    if exp_setings.resume:
        flags.append('resume')
    if exp_setings.continue_with_errors:
        flags.append('continue_with_errors')
    return flags


class Interface:
    def __init__(self, parameter_file: str = None, exp_settings: ExpSettings = None, share: bool = False):
        with gr.Blocks() as demo:
            gr.Markdown(TITLE)
            with gr.Tab("Training"):
                TrainingInterface(parameter_file, exp_settings)
            with gr.Tab("Inference"):
                InferenceInterface()

        demo.queue()
        demo.launch(share=share)


class InferenceInterface:
    def __init__(self):
        use_gpu = gr.components.Checkbox(label="Use GPU")
        with gr.Row():
            model = gr.components.File(type='file', label="Model checkpoint in Wandb folder")
            config = gr.components.File(type='file', label="Wandb config input")
        text = gr.Textbox(TEXT, show_label=False)
        divider = gr.Markdown(DIV)
        with gr.Row() as self.input_images_row:
            self.input_channels = {
                'R': gr.Image(type='pil', image_mode='L', label='R Channel', visible=False),
                'G': gr.Image(type='pil', image_mode='L', label='G Channel', visible=False),
                'B': gr.Image(type='pil', image_mode='L', label='B Channel', visible=False),
                "NDVI": gr.Image(type='pil', image_mode='L', label='NDVI Channel', visible=False),
                "NIR": gr.Image(type='pil', image_mode='L', label='NIR Channel', visible=False),
            }
        config.change(self.set_model,
                      inputs=[model, config, use_gpu],
                      outputs=[*self.input_channels.values(), text],
                      )
        model.change(self.set_model,
                     inputs=[model, config, use_gpu],
                     outputs=[*self.input_channels.values(), text],
                     )
        submit = gr.Button(variant="primary")
        segmentation = gr.Image(label="Segmentation")

        submit.click(self.predict, inputs=[use_gpu, *self.input_channels.values()], outputs=[segmentation])

    def set_model(self, model_path_wrapper, config_wrapper, gpu):
        if model_path_wrapper is None or config_wrapper is None:
            return [gr.update(visible=False) for _ in self.input_channels] + [TEXT]
        self.inferencer = WandbInferencer(model_path_wrapper, config_wrapper, gpu)
        input_channels = []
        for channel, input_form in self.input_channels.items():
            if channel in self.inferencer.channels:
                input_channels.append(gr.update(visible=True))
            else:
                input_channels.append(gr.update(visible=False))
        return input_channels + ["".join([f"{channel} " for channel in self.inferencer.channels]) + "needed"]

    def predict(self, use_gpu, *inputs):
        if use_gpu:
            self.inferencer.cuda()
        else:
            self.inferencer.cpu()
        inputs = filter(lambda x: x is not None, inputs)
        image = compose_images(list(inputs))
        pred = self.inferencer(image)
        segmentation = segment(pred)
        return segmentation


class TrainingInterface:

    def __init__(self, parameter_file: str = None, exp_settings: ExpSettings = None):
        self.experimenter = Experimenter()
        default_updates = self.set_file(parameter_file)
        self.experimenter.exp_settings.update(exp_settings)
        exp_settings = self.experimenter.exp_settings
        gr.Markdown("Train your model!")
        with gr.Row():
            self.group_txt = gr.Textbox(label="Experiment group", value=exp_settings.group)
            self.track_dir = gr.Textbox(label="Tracking directory", value=exp_settings.tracking_dir)
            self.start_from_grid_txt = gr.Number(label="Start from grid",
                                                 precision=0,
                                                 value=exp_settings.start_from_grid)
            self.start_from_run_txt = gr.Number(label="Start from run",
                                                precision=0,
                                                value=exp_settings.start_from_run)

            self.flags = gr.CheckboxGroup(["resume", "continue_with_errors"], label="", value=flag_maker(exp_settings))

        self.parameters = gr.File(label="Parameter file", value=parameter_file)
        self.group_mk = gr.Markdown(exp_settings.group)
        with gr.Row():
            self.exp_summary = gr.Markdown(value=default_updates["exp_summary"])
            self.grids_dataframe = gr.Markdown(value=default_updates["grids_dataframe"])

        parameters_comps = [self.group_txt,
                            self.track_dir, self.start_from_grid_txt,
                            self.start_from_run_txt, self.flags]
        for com in parameters_comps:
            com.change(self.update_from_gui, inputs=parameters_comps, outputs=[self.group_mk, self.exp_summary])
        self.parameters.change(self.update_from_file, inputs=[self.parameters],
                               outputs=[self.group_mk, self.exp_summary, self.grids_dataframe] + parameters_comps)

        self.experiment_btn = gr.Button("Experiment!", variant="primary")
        self.progress = gr.Label(value={}, label="Progress")
        with gr.Row():
            self.json = gr.JSON(label="Current run params")
            self.failures = MkFailures()
            self.failures_mk = gr.Markdown(self.failures.get_text())
        self.experiment_btn.click(self.experiment, inputs=[], outputs=[self.progress, self.json, self.failures_mk])

    def experiment(self):
        yield self.update_progress(self.experimenter.exp_settings.start_from_grid,
                                   self.experimenter.exp_settings.start_from_run - 1,
                                   len(self.experimenter.grids),
                                   len(self.experimenter.grids[self.experimenter.exp_settings.start_from_grid]),
                                   "starting",
                                   {})
        for out in self.experimenter.execute_runs_generator(callback=self.update_progress):
            yield out

    def update_progress(self, cur_grid, cur_run, n_grids, n_runs, status, run_params, exception=None):
        d = {}
        for i in range(n_grids):
            if i < cur_grid:
                d[f"Grid {i}"] = 1
            elif i == cur_grid:
                if status in ["finished", "crashed"]:
                    cur_run += 1
                d[f"Grid {i} ({cur_run} / {n_runs})"] = cur_run / n_runs
            else:
                d[f"Grid {i}"] = 0
        if status == "crashed":
            self.failures.update(cur_grid, cur_run, exception)
        return d, run_params, self.failures.get_text()

    def set_file(self, file):
        self.experimenter = Experimenter()
        self.experimenter.exp_settings = ExpSettings()
        if file is None:
            return DEFAULTS([
                "group",
                "track_dir",
                "start_from_grid",
                "start_from_run",
                "flags",
                "group_mk",
                "grids_dataframe",
                "exp_summary",
            ])

        settings = load_yaml(file.name)
        sm, grids, dots = self.experimenter.calculate_runs(settings)
        sum_mk = exp_summary_builder(self.experimenter)
        params_mk = grid_summary_builder(grids, dots)
        flags = flag_maker(self.experimenter.exp_settings)
        return {
                "group": self.experimenter.exp_settings['group'],
                "track_dir": self.experimenter.exp_settings['tracking_dir'] or DEFAULTS('tracking_dir'),
                "start_from_grid": self.experimenter.exp_settings['start_from_grid'],
                "start_from_run": self.experimenter.exp_settings['start_from_run'],
                "flags": flags,
                "group_mk": title_builder(self.experimenter.exp_settings['group']),
                "exp_summary": sum_mk,
                "grids_dataframe": params_mk
            }

    def update_from_file(self, file):
        updates = self.set_file(file)
        return {
                self.group_txt: self.experimenter.exp_settings['group'],
                self.track_dir: self.experimenter.exp_settings['tracking_dir'] or DEFAULTS('tracking_dir'),
                self.start_from_grid_txt: self.experimenter.exp_settings['start_from_grid'],
                self.start_from_run_txt: self.experimenter.exp_settings['start_from_run'],
                self.flags: updates["flags"],
                self.group_mk: title_builder(self.experimenter.exp_settings['group']),
                self.exp_summary: updates["exp_summary"],
                self.grids_dataframe: updates["grids_dataframe"]
            }

    def update_from_gui(self, name, track_dir, start_grid, start_run, flags):
        if self.experimenter is None:
            self.experimenter = Experimenter()
        self.experimenter.update_settings({
            "resume": ("resume" in flags),
            "continue_with_errors": ("continue_with_errors" in flags),
            "start_from_grid": start_grid,
            "start_from_run": start_run,
            "tracking_dir": track_dir,
            "group": name,
        })
        return title_builder(self.experimenter.exp_settings['group']) or "", exp_summary_builder(self.experimenter)


def frontend(parameter_file, args, share):
    settings = ExpSettings(**args)
    Interface(parameter_file, settings, share)



from typing import Union

import wandb
from ruamel.yaml import YAML

from ezdl.utils.utilities import values_to_number, nested_dict_update
from tqdm import tqdm


def fix_string_param(run):
    try:
        params = values_to_number(run.config['in_params'])
        run.config['in_params'] = params
    except KeyError:
        print(f'{run.name} has no in_params')
    run.update()


def update_config(run, config):
    try:
        params = run.config['in_params']
        params = nested_dict_update(params, config)
        run.config['in_params'] = params
    except KeyError:
        print(f'{run.name} has no config')
    run.update()


def remove_key(run, config):
    try:
        params = run.config['in_params']
        for key in config:
            params.pop(key)
        run.config['in_params'] = params
    except KeyError:
        print(f'{run.name} has no config')
    run.update()


def update_metadata(run, metadata):
    for key, value in metadata.items():
        try:
            run.__dict__[key] = value
        except KeyError:
            print(f'{run.name} has no {key}')
    run.update()


def delete_files(run, files_to_delete: Union[list, str]):
    if type(files_to_delete) == list:
        to_delete = files_to_delete.copy()
        files = run.files()
        for file in files:
            if file.name in to_delete:
                to_delete.pop(to_delete.index(file.name))
                file.delete()
                print(f"{file.name} deleted")
            if len(to_delete) == 0:
                print('All files deleted')
                return
            print(f"{file.name} deleted")

    elif type(files_to_delete) == str:
        to_delete = files_to_delete
        files = run.files()
        for file in files:
            if eval(to_delete)(file):
                file.delete()
                print(f"{file.name} deleted")
    else:
        raise NotImplemented("Use lists or strings")
    print(f"Files remained: {to_delete}")


def delete_artifacts(run, artifacts_to_delete: Union[list, str]):
    if type(artifacts_to_delete) == list:
        to_delete = artifacts_to_delete.copy()
        artifacts = run.logged_artifacts()
        for art in artifacts:
            if art.name in to_delete:
                to_delete.pop(to_delete.index(art.name))
                art.delete()
                print(f"{art.name} deleted")
            if len(to_delete) == 0:
                print('All files deleted')
                return
            print(f"{art.name} deleted")

    elif type(artifacts_to_delete) == str:
        to_delete = artifacts_to_delete
        artifacts = run.logged_artifacts()
        for art in artifacts:
            if eval(to_delete)(art):
                art.delete(delete_aliases=True)
                print(f"{art.name} deleted")
    else:
        raise NotImplemented("Use lists or strings")
    print(f"Artifacts remained: {to_delete}")


def manipulate():
    param_path = 'manip.yaml'
    with open(param_path, 'r') as param_stream:
        settings = YAML().load(param_stream)
    queries = settings['runs']
    path = settings['path']
    for query in queries:
        print('Query:', query)
        filters = query['filters']
        fix_string_params = query['fix_string_params']
        updated_config = query['updated_config']
        updated_metadata = query['updated_meta']
        keys_to_delete = query['keys_to_delete']
        files_to_delete = query['files_to_delete']
        artifacts_to_delete = query['artifacts_to_delete']
        api = wandb.Api()
        runs = api.runs(path=path, filters=filters)
        if len(runs) != 0:
            for run in tqdm(runs):
                print(f"Name: {run.name} Id: {run.id}")
                if fix_string_params:
                    fix_string_param(run)
                if updated_config is None:
                    print('No config to update')
                else:
                    update_config(run, updated_config)
                if updated_metadata is None:
                    print('No metadata to update')
                else:
                    update_metadata(run, updated_metadata)
                if keys_to_delete is None:
                    print('No keys to delete')
                else:
                    remove_key(run, keys_to_delete)
                if files_to_delete is None:
                    print("No files to delete")
                else:
                    delete_files(run, files_to_delete)
                if artifacts_to_delete is None:
                    print("No artifacts to delete")
                else:
                    delete_artifacts(run, artifacts_to_delete)




import os
from typing import Optional, Union

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import adjectiveanimalnumber as aan
import torch

from PIL import Image
from flatbuffers.builder import np
from matplotlib import pyplot as plt
from super_gradients.common.abstractions.abstract_logger import get_logger
from super_gradients.common.environment.ddp_utils import multi_process_safe
from clearml import Task, OutputModel

from ezdl.logger.basesg_logger import BaseSGLogger
from ezdl.utils.segmentation import tensor_to_segmentation_image

logger = get_logger(__name__)

WANDB_ID_PREFIX = 'wandb_id.'
WANDB_INCLUDE_FILE_NAME = '.wandbinclude'


class ClearMLLogger(BaseSGLogger):

    def __init__(self, project_name: str, experiment_name: str, storage_location: str, resumed: bool,
                 training_params: dict, checkpoints_dir_path: str, tb_files_user_prompt: bool = False,
                 launch_tensorboard: bool = False, tensorboard_port: int = None, save_checkpoints_remote: bool = True,
                 save_tensorboard_remote: bool = True, save_logs_remote: bool = True, entity: Optional[str] = None,
                 api_server: Optional[str] = None, save_code: bool = False, tags=None, run_id=None, **kwargs):
        """

        :param experiment_name: Used for logging and loading purposes
        :param s3_path: If set to 's3' (i.e. s3://my-bucket) saves the Checkpoints in AWS S3 otherwise saves the Checkpoints Locally
        :param checkpoint_loaded: if true, then old tensorboard files will *not* be deleted when tb_files_user_prompt=True
        :param max_epochs: the number of epochs planned for this training
        :param tb_files_user_prompt: Asks user for Tensorboard deletion prompt.
        :param launch_tensorboard: Whether to launch a TensorBoard process.
        :param tensorboard_port: Specific port number for the tensorboard to use when launched (when set to None, some free port
                    number will be used
        :param save_checkpoints_remote: Saves checkpoints in s3.
        :param save_tensorboard_remote: Saves tensorboard in s3.
        :param save_logs_remote: Saves log files in s3.
        :param save_code: save current code to wandb
        """
        self.s3_location_available = storage_location.startswith('s3')
        self.resumed = resumed
        resume = 'must' if resumed else None
        os.makedirs(checkpoints_dir_path, exist_ok=True)
        if kwargs.get("group"):
            project_name = f"{project_name}/{kwargs.get('group')}"
        if not experiment_name:
            experiment_name = aan.generate()

        self.run = Task.init(project_name=project_name,
                             task_name=experiment_name,
                             auto_connect_frameworks=False,
                             continue_last_task=resume,
                             )
        if save_code:
            self._save_code()

        self.save_checkpoints_wandb = save_checkpoints_remote
        self.save_tensorboard_wandb = save_tensorboard_remote
        super().__init__(project_name, experiment_name, storage_location, resumed, training_params,
                         checkpoints_dir_path, tb_files_user_prompt, launch_tensorboard, tensorboard_port,
                         self.s3_location_available, self.s3_location_available, self.s3_location_available)

    @multi_process_safe
    def add_config(self, tag: str = None, config: dict = None):
        self.run.connect(config, name=tag)

    @multi_process_safe
    def add_scalar(self, tag: str, scalar_value: float, global_step: int = 0):
        self.run.get_logger().report_scalar(title=tag, series=tag, value=scalar_value, iteration=global_step)

    @multi_process_safe
    def add_scalars(self, tag_scalar_dict: dict, global_step: int = 0):
        for name, value in tag_scalar_dict.items():
            self.add_scalar(name, value, global_step=global_step)

    @multi_process_safe
    def add_image(self, tag: str, image: Union[torch.Tensor, np.array, Image.Image], data_format='CHW', global_step: int = 0):
        if isinstance(image, torch.Tensor):
            image = image.cpu().detach().numpy()
        if image.shape[0] < 5:
            image = image.transpose([1, 2, 0])
        self.run.get_logger().report_image(title=tag, series=tag, image=image, iteration=global_step)

    @multi_process_safe
    def add_images(self, tag: str, images: Union[torch.Tensor, np.array], data_format='NCHW', global_step: int = 0):
        raise NotImplementedError()


    @multi_process_safe
    def add_video(self, tag: str, video: Union[torch.Tensor, np.array], global_step: int = 0):
        raise NotImplementedError()

    @multi_process_safe
    def add_histogram(self, tag: str, values: Union[torch.Tensor, np.array], bins: str, global_step: int = 0):
        self.run.get_logger().report_histogram(title=tag, series=tag, values=values, iteration=global_step)

    @multi_process_safe
    def add_text(self, tag: str, text_string: str, global_step: int = 0):
        self.run.get_logger().report_text(msg=text_string)

    @multi_process_safe
    def add_figure(self, tag: str, figure: plt.figure, global_step: int = 0):
        self.run.get_logger().report_matplotlib_figure(title=tag, figure=figure, iteration=global_step)

    @multi_process_safe
    def add_table(self, tag, data, columns, rows):
        self.run.get_logger().report_table(title=tag, series=tag, table_plot=pd.DataFrame(data, columns=columns))

    @multi_process_safe
    def add_plot(self, tag: str, values: pd.DataFrame, xtitle, ytitle, classes_marker=None):

        if classes_marker:
            for cls in values[classes_marker].unique():
                scatter2d = values[values[classes_marker] == cls].loc[:, values.columns != classes_marker].values

                self.run.get_logger().current_logger().report_scatter2d(
                    tag,
                    tag,
                    iteration=None,
                    scatter=scatter2d,
                    xaxis=xtitle,
                    yaxis=ytitle,
                    mode='lines'
                )
        else:
            self.run.get_logger().current_logger().report_scatter2d(
                tag,
                tag,
                iteration=None,
                scatter=values.values,
                xaxis=xtitle,
                yaxis=ytitle,
                mode='lines+markers'
            )

    @multi_process_safe
    def close(self, really=False):
        if really:
            super().close()
            self.run.close()

    @multi_process_safe
    def add_file(self, file_name: str = None):
        self.run.upload_artifact(file_name, artifact_object=os.path.join(self._local_dir, file_name))

    @multi_process_safe
    def add_summary(self, metrics: dict):
        for name, value in metrics.items():
            self.run.get_logger().report_single_value(name=name, value=value)

    @multi_process_safe
    def add_checkpoint(self, tag: str, state_dict: dict, global_step: int = 0):
        name = 'ckpt.pth' if tag is None else tag
        if not name.endswith('.pth'):
            name += '.pth'
        model = OutputModel(task=self.run, name=name)
        path = os.path.join(self._local_dir, name)
        torch.save(state_dict, path)
        model.update_weights(weights_filename=path, iteration=global_step)

    @multi_process_safe
    def add_mask(self, tag: str, image, mask_dict, global_step: int = 0):
        predictions = tensor_to_segmentation_image(mask_dict['predictions']['mask_data'],
                                                   labels=mask_dict['predictions']['class_labels'])
        ground_truth, cmap = tensor_to_segmentation_image(mask_dict['ground_truth']['mask_data'],
                                                    labels=mask_dict['ground_truth']['class_labels'], return_cmap=True)
        cmap = {name: '#%02x%02x%02x' % (c[0], c[1], c[2]) for name, c in cmap.items()}
        data = np.stack([image, predictions, ground_truth])
        fig = px.imshow(data, facet_col=0, title=tag)
        annotations = ["image", "predictions", "ground_truth"]
        for k in range(len(annotations)):
            fig.layout.annotations[k].update(text=annotations[k])
        fig.add_traces([
            go.Scatter(x=[None], y=[None], name=name, mode='markers', marker=dict(color=color, size=1))
            for name, color in cmap.items()
        ])
        self.run.get_logger().report_plotly(
            title=tag, series=tag, iteration=global_step, figure=fig
        )

    def _get_tensorboard_file_name(self):
        try:
            tb_file_path = self.tensorboard_writer.file_writer.event_writer._file_name
        except RuntimeError as e:
            logger.warning('tensorboard file could not be located for ')
            return None

        return tb_file_path

    @staticmethod
    def _search_upwards_for_file(file_name: str):
        """
        Search in the current directory and all directories above it for a file of a particular name.
        :param file_name: file name to look for.
        :return: pathlib.Path, the location of the first file found or None, if none was found
        """

        try:
            cur_dir = os.getcwd()
            while cur_dir != '/':
                if file_name in os.listdir(cur_dir):
                    return os.path.join(cur_dir, file_name)
                else:
                    cur_dir = os.path.dirname(cur_dir)
        except RuntimeError as e:
            return None

        return None

    def create_image_mask_sequence(self, name):
        pass

    def add_image_mask_to_sequence(self, sequence_name, name, image, mask_dict):
        self.add_mask(name, image, mask_dict)

    def add_image_mask_sequence(self, name):
        pass

    def __repr__(self):
        return "ClearMLLogger"



import torch
from torch import nn, Tensor
from typing import Tuple
from torch.nn import functional as F


class MLP(nn.Module):
    def __init__(self, dim, embed_dim):
        super().__init__()
        self.proj = nn.Linear(dim, embed_dim)

    def forward(self, x: Tensor) -> Tensor:
        x = x.flatten(2).transpose(1, 2)
        x = self.proj(x)
        return x


class ConvModule(nn.Module):
    def __init__(self, c1, c2):
        super().__init__()
        self.conv = nn.Conv2d(c1, c2, 1, bias=False)
        self.bn = nn.BatchNorm2d(c2)        # use SyncBN in original
        self.activate = nn.ReLU(True)

    def forward(self, x: Tensor) -> Tensor:
        return self.activate(self.bn(self.conv(x)))


class SegFormerHead(nn.Module):
    def __init__(self, dims: list, embed_dim: int = 256, num_classes: int = 19):
        super().__init__()
        for i, dim in enumerate(dims):
            self.add_module(f"linear_c{i+1}", MLP(dim, embed_dim))

        self.linear_fuse = ConvModule(embed_dim*4, embed_dim)
        self.linear_pred = nn.Conv2d(embed_dim, num_classes, 1)
        self.dropout = nn.Dropout2d(0.1)

    def forward(self, features: Tuple[Tensor, Tensor, Tensor, Tensor]) -> Tensor:
        B, _, H, W = features[0].shape
        outs = [self.linear_c1(features[0]).permute(0, 2, 1).reshape(B, -1, *features[0].shape[-2:])]

        for i, feature in enumerate(features[1:]):
            cf = eval(f"self.linear_c{i+2}")(feature).permute(0, 2, 1).reshape(B, -1, *feature.shape[-2:])
            outs.append(F.interpolate(cf, size=(H, W), mode='bilinear', align_corners=False))

        seg = self.linear_fuse(torch.cat(outs[::-1], dim=1))
        seg = self.linear_pred(self.dropout(seg))
        return seg